"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        if (typeof b !== "function" && b !== null)
            throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
exports.__esModule = true;
exports.Result = exports.DestroyableParentPath = exports.ParentPath = exports.HasCapacity = exports.HasLength = exports.EMPTY_OBJECT = exports.EMPTY_ARRAY = exports.mapImplemented = exports.noop = void 0;
var iterables_1 = require("./iterables");
var value_1 = require("./value");
/**
 * Does nothing.
 */
var noop = function () { };
exports.noop = noop;
/**
 * `true` if `Map` is implemented.
 */
var mapImplemented = function () {
    return Boolean(Map && Map.prototype && typeof Map.prototype.entries === "function");
};
exports.mapImplemented = mapImplemented;
/**
 * An empty array. MUST NOT BE MODIFIED.
 */
exports.EMPTY_ARRAY = [];
/**
 * An empty object. MUST NOT BE MODIFIED.
 */
exports.EMPTY_OBJECT = {};
/**
 * Describes a container having a length.
 */
var HasLength = /** @class */ (function () {
    function HasLength() {
    }
    /**
     * Returns `true` if the underlying container is empty (has zero length).
     */
    HasLength.prototype.isEmpty = function () {
        return this.len() === 0;
    };
    return HasLength;
}());
exports.HasLength = HasLength;
/**
 * Describes a container having both capacity and length.
 */
var HasCapacity = /** @class */ (function (_super) {
    __extends(HasCapacity, _super);
    /**
     * Creates a new container with supplied capacity.
     * In case capacity is equal to zero, an error will be thrown.
     * @param capacity
     */
    function HasCapacity(capacity) {
        var _this = _super.call(this) || this;
        if (capacity === 0) {
            throw new Error("Capacity can't be equal to zero");
        }
        _this._capacity = capacity;
        return _this;
    }
    /**
     * Returns capacity of the underlying container.
     *
     *
     */
    HasCapacity.prototype.capacity = function () {
        return this._capacity;
    };
    /**
     * Returns `true` if the length of the underlying container reached its capacity.
     *
     *
     */
    HasCapacity.prototype.isFull = function () {
        return this.capacity() < this.len();
    };
    /**
     * Returns `true` if the length of the underlying container will reach (or already reached) its capacity.
     *
     *
     */
    HasCapacity.prototype.willBeFull = function () {
        return this.capacity() < this.len() + 1;
    };
    return HasCapacity;
}(HasLength));
exports.HasCapacity = HasCapacity;
/**
 * The path from the parent to the child.
 * If the key is a `NO_VALUE`, then the child is stored under the key equal to itself.
 */
var ParentPath = /** @class */ (function () {
    function ParentPath(parent, key) {
        this.parent = parent;
        this.key = key;
    }
    /// Enforces parent to drop the supplied child. Its value will be used only in case if key is absent.
    ParentPath.prototype.drop = function (child) {
        this.parent.drop(this.key === value_1.NO_VALUE ? child : this.key);
    };
    return ParentPath;
}());
exports.ParentPath = ParentPath;
/** Same as `ParentPath` but will destroy the parent if it became empty. */
var DestroyableParentPath = /** @class */ (function (_super) {
    __extends(DestroyableParentPath, _super);
    function DestroyableParentPath() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    /// Enforces parent to drop the supplied child. Its value will be used only in case if key is absent.
    DestroyableParentPath.prototype.drop = function (child) {
        _super.prototype.drop.call(this, child);
        if (this.parent.isEmpty()) {
            this.parent.destroy();
        }
    };
    return DestroyableParentPath;
}(ParentPath));
exports.DestroyableParentPath = DestroyableParentPath;
/**
 * Contains removed/added entities.
 */
var Result = /** @class */ (function () {
    function Result(removed, added) {
        if (removed === void 0) { removed = iterables_1.EMPTY_ITERABLE; }
        if (added === void 0) { added = iterables_1.EMPTY_ITERABLE; }
        this.removed = removed;
        this.added = added;
    }
    /** Creates a result containing added entities */
    Result.added = function (added) {
        return new Result(void 0, added);
    };
    /** Creates a result containing removed entities */
    Result.removed = function (removed) {
        return new Result(removed);
    };
    /**
     * Creates a result containing removed and added entities.
     * @param removed
     * @param added
     */
    Result.removedAdded = function (removed, added) {
        return new Result(removed, added);
    };
    /**
     * Creates an empty result.
     */
    Result.empty = function () {
        return Result.EMPTY_RESULT;
    };
    /**
     * Appends remove/added items from the supplied result to the current result.
     * @param result
     */
    Result.prototype.chain = function (result) {
        if (result === Result.EMPTY_RESULT) {
            return this;
        }
        else if (this === Result.EMPTY_RESULT) {
            return result;
        }
        return new Result((0, iterables_1.chain)(this.removed, result.removed), (0, iterables_1.chain)(this.added, result.added));
    };
    /**
     * Executes given function for each added item.
     * @param fn
     *
     */
    Result.prototype.forEachAdded = function (fn) {
        (0, iterables_1.forEach)(fn, this.added);
        return this;
    };
    /**
     * Executes given function for each removed item.
     * @param fn
     *
     */
    Result.prototype.forEachRemoved = function (fn) {
        (0, iterables_1.forEach)(fn, this.removed);
        return this;
    };
    Result.EMPTY_RESULT = new Result();
    return Result;
}());
exports.Result = Result;
//# sourceMappingURL=utils.js.map